Activities Resource (/activities)
The Activities Resource provides a method to retrieve a full description of an Activity from the LRS.
Any deviation from section 4.1.6.4 Activities Resource (/activities) of the xAPI specification is a bug.