Agents Resource (/agents)
Provides a method to retrieve a Person Object with combined information about an Agent derived from an outside service, such as a directory service. This object is called a “Person Object”. This Person Object is very similar to an Agent Object, but instead of each attribute having a single value, each attribute has an array value. In addition it’s legal to include multiple identifying properties.
Any deviation from section 4.1.6.3 Agents Resource (/activities/state) of the xAPI specification is a bug.