Agent Profile Resource (/agents/profile)
A place to store information about an Agent or an Identified Group
in a generic form called a document. This information is not tied to an
activity or registration. The semantics of the LRS response are driven
by the presence of a profileId parameter. If it is included, the GET
and DELETE methods acts upon a single defined profile document
identified by profileId. Otherwise, GET` returns the available ids
given through the other parameter.
IMPORTANT - This resource has concurrency controls associated w/ it.
Any deviation from section 4.1.6.5 Agent Profile Resource (/agents/profile) of the xAPI specification is a bug.